"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.assertInstructionNode = exports.isInstructionNode = exports.InstructionNode = void 0;
const utils_1 = require("../utils");
const TypeNode_1 = require("./TypeNode");
const TypeStructNode_1 = require("./TypeStructNode");
const TypeStructFieldNode_1 = require("./TypeStructFieldNode");
const ValueNode_1 = require("./ValueNode");
class InstructionNode {
    constructor(metadata, accounts, args, subInstructions) {
        this.nodeClass = 'InstructionNode';
        const bytes = metadata.bytesCreatedOnChain;
        this.metadata = {
            ...metadata,
            name: (0, utils_1.mainCase)(metadata.name),
            bytesCreatedOnChain: 'name' in bytes ? { ...bytes, name: (0, utils_1.mainCase)(bytes.name) } : bytes,
        };
        this.accounts = accounts.map((account) => {
            const { defaultsTo } = account;
            if (defaultsTo.kind === 'account') {
                defaultsTo.name = (0, utils_1.mainCase)(defaultsTo.name);
            }
            else if (defaultsTo.kind === 'program') {
                defaultsTo.program.name = (0, utils_1.mainCase)(defaultsTo.program.name);
            }
            else if (defaultsTo.kind === 'pda') {
                defaultsTo.pdaAccount = (0, utils_1.mainCase)(defaultsTo.pdaAccount);
                defaultsTo.seeds = Object.fromEntries(Object.entries(defaultsTo.seeds).map(([key, seed]) => [
                    (0, utils_1.mainCase)(key),
                    { ...seed, name: (0, utils_1.mainCase)(seed.name) },
                ]));
            }
            return { ...account, name: (0, utils_1.mainCase)(account.name), defaultsTo };
        });
        this.args = args;
        this.subInstructions = subInstructions;
    }
    static fromIdl(idl) {
        var _a, _b, _c, _d, _e;
        const idlName = (_a = idl.name) !== null && _a !== void 0 ? _a : '';
        const name = (0, utils_1.camelCase)(idlName);
        const useProgramIdForOptionalAccounts = (_b = idl.defaultOptionalAccounts) !== null && _b !== void 0 ? _b : false;
        const metadata = {
            name,
            idlName,
            docs: (_c = idl.docs) !== null && _c !== void 0 ? _c : [],
            internal: false,
            bytesCreatedOnChain: { kind: 'none' },
        };
        const accounts = ((_d = idl.accounts) !== null && _d !== void 0 ? _d : []).map((account) => {
            var _a, _b, _c, _d, _e, _f;
            const isOptional = (_a = account.optional) !== null && _a !== void 0 ? _a : false;
            return {
                name: (0, utils_1.camelCase)((_b = account.name) !== null && _b !== void 0 ? _b : ''),
                isWritable: (_c = account.isMut) !== null && _c !== void 0 ? _c : false,
                isSigner: (_d = account.isSigner) !== null && _d !== void 0 ? _d : false,
                isOptionalSigner: (_e = account.isOptionalSigner) !== null && _e !== void 0 ? _e : false,
                isOptional,
                description: (_f = account.desc) !== null && _f !== void 0 ? _f : '',
                defaultsTo: isOptional && useProgramIdForOptionalAccounts
                    ? { kind: 'programId' }
                    : { kind: 'none' },
            };
        });
        let args = TypeStructNode_1.TypeStructNode.fromIdl({
            kind: 'struct',
            name: name ? `${name}InstructionArgs` : '',
            fields: (_e = idl.args) !== null && _e !== void 0 ? _e : [],
        });
        if (idl.discriminant) {
            const discriminatorField = new TypeStructFieldNode_1.TypeStructFieldNode({
                name: 'discriminator',
                docs: [],
                defaultsTo: {
                    strategy: 'omitted',
                    value: (0, ValueNode_1.vScalar)(idl.discriminant.value),
                },
            }, (0, TypeNode_1.createTypeNodeFromIdl)(idl.discriminant.type));
            args = new TypeStructNode_1.TypeStructNode(args.name, [
                discriminatorField,
                ...args.fields,
            ]);
        }
        return new InstructionNode(metadata, accounts, args, []);
    }
    accept(visitor) {
        return visitor.visitInstruction(this);
    }
    getAllSubInstructions() {
        return this.subInstructions.flatMap((subInstruction) => [
            subInstruction,
            ...subInstruction.getAllSubInstructions(),
        ]);
    }
    get name() {
        return this.metadata.name;
    }
    get docs() {
        return this.metadata.docs;
    }
    get hasAccounts() {
        return this.accounts.length > 0;
    }
    get hasData() {
        return this.args.fields.length > 0;
    }
    get hasArgs() {
        const nonOmittedFields = this.args.fields.filter((field) => { var _a; return ((_a = field.metadata.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; });
        return nonOmittedFields.length > 0;
    }
    get hasRequiredArgs() {
        const requiredFields = this.args.fields.filter((field) => field.metadata.defaultsTo === null);
        return requiredFields.length > 0;
    }
}
exports.InstructionNode = InstructionNode;
function isInstructionNode(node) {
    return !!node && node.nodeClass === 'InstructionNode';
}
exports.isInstructionNode = isInstructionNode;
function assertInstructionNode(node) {
    var _a;
    if (!isInstructionNode(node)) {
        throw new Error(`Expected InstructionNode, got ${(_a = node === null || node === void 0 ? void 0 : node.nodeClass) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertInstructionNode = assertInstructionNode;
//# sourceMappingURL=InstructionNode.js.map