"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetByteSizeVisitor = void 0;
const BaseThrowVisitor_1 = require("../BaseThrowVisitor");
class GetByteSizeVisitor extends BaseThrowVisitor_1.BaseThrowVisitor {
    constructor() {
        super(...arguments);
        this.availableDefinedTypes = new Map();
        this.visitedDefinedTypes = new Map();
        this.definedTypeStack = [];
    }
    registerDefinedTypes(definedTypes) {
        definedTypes.forEach((definedType) => {
            this.availableDefinedTypes.set(definedType.name, definedType);
        });
    }
    visitAccount(account) {
        return account.type.accept(this);
    }
    visitInstruction(instruction) {
        return instruction.args.accept(this);
    }
    visitDefinedType(definedType) {
        if (this.visitedDefinedTypes.has(definedType.name)) {
            return this.visitedDefinedTypes.get(definedType.name);
        }
        this.definedTypeStack.push(definedType.name);
        const child = definedType.type.accept(this);
        this.definedTypeStack.pop();
        this.visitedDefinedTypes.set(definedType.name, child);
        return child;
    }
    visitTypeArray(typeArray) {
        const itemSize = typeArray.itemType.accept(this);
        const arraySize = itemSize !== null ? itemSize * typeArray.size : null;
        return typeArray.size === 0 ? 0 : arraySize;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        const linkedDefinedType = this.availableDefinedTypes.get(typeDefinedLink.definedType);
        if (!linkedDefinedType) {
            throw new Error(`Cannot find linked defined type ${typeDefinedLink.definedType}.`);
        }
        // This prevents infinite recursion by using assuming
        // cyclic types don't have a fixed size.
        if (this.definedTypeStack.includes(linkedDefinedType.name)) {
            return null;
        }
        return linkedDefinedType.accept(this);
    }
    visitTypeEnum(typeEnum) {
        if (typeEnum.isScalarEnum())
            return 1;
        const variantSizes = typeEnum.variants.map((v) => v.accept(this));
        const allVariantHaveTheSameFixedSize = variantSizes.every((one, i, all) => one === all[0]);
        return allVariantHaveTheSameFixedSize &&
            variantSizes.length > 0 &&
            variantSizes[0] !== null
            ? variantSizes[0] + 1
            : null;
    }
    visitTypeEnumEmptyVariant() {
        return 0;
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        return typeEnumStructVariant.struct.accept(this);
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        return typeEnumTupleVariant.tuple.accept(this);
    }
    visitTypeLeaf(typeLeaf) {
        switch (typeLeaf.type) {
            case 'bool':
            case 'u8':
            case 'i8':
                return 1;
            case 'u16':
            case 'i16':
                return 2;
            case 'u32':
            case 'i32':
            case 'f32':
                return 4;
            case 'u64':
            case 'i64':
            case 'f64':
                return 8;
            case 'u128':
            case 'i128':
                return 16;
            case 'publicKey':
                return 32;
            case 'string':
            case 'bytes':
            default:
                return null;
        }
    }
    visitTypeLeafWrapper(typeLeafWrapper) {
        return typeLeafWrapper.leaf.accept(this);
    }
    visitTypeMap() {
        return null;
    }
    visitTypeOption(typeOption) {
        const child = typeOption.type.accept(this);
        if (typeOption.optionType === 'option')
            return null;
        return child !== null ? child + 4 : null;
    }
    visitTypeSet() {
        return null;
    }
    visitTypeStruct(typeStruct) {
        return this.sumSizes(typeStruct.fields.map((f) => f.accept(this)));
    }
    visitTypeStructField(typeStructField) {
        return typeStructField.type.accept(this);
    }
    visitTypeTuple(typeTuple) {
        return this.sumSizes(typeTuple.itemTypes.map((i) => i.accept(this)));
    }
    visitTypeVec() {
        return null;
    }
    sumSizes(sizes) {
        return sizes.reduce((all, one) => (all === null || one === null ? null : all + one), 0);
    }
}
exports.GetByteSizeVisitor = GetByteSizeVisitor;
//# sourceMappingURL=GetByteSizeVisitor.js.map