"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetNodeInlineStringVisitor = void 0;
const ROOT_PREFIX = 'R';
const PROGRAM_PREFIX = 'P';
const ACCOUNT_PREFIX = 'A';
const INSTRUCTION_PREFIX = 'I';
const TYPE_PREFIX = 'T';
const ERROR_PREFIX = 'E';
class GetNodeInlineStringVisitor {
    visitRoot(root) {
        const children = root.programs.map((program) => program.accept(this));
        return `${ROOT_PREFIX}(${children.join(',')})`;
    }
    visitProgram(program) {
        const children = [
            ...program.accounts.map((account) => account.accept(this)),
            ...program.instructionsWithSubs.map((ix) => ix.accept(this)),
            ...program.definedTypes.map((type) => type.accept(this)),
            ...program.errors.map((type) => type.accept(this)),
        ];
        return `${PROGRAM_PREFIX}[${program.name}](${children.join(',')})`;
    }
    visitAccount(account) {
        const child = account.type.accept(this);
        return `${ACCOUNT_PREFIX}[${account.name}](${child})`;
    }
    visitInstruction(instruction) {
        const accounts = instruction.accounts.map((account) => account.name);
        const args = instruction.args.accept(this);
        return (`${INSTRUCTION_PREFIX}[${instruction.name}](` +
            `accounts:(${accounts.join(',')}),` +
            `args:(${args})` +
            `)`);
    }
    visitDefinedType(definedType) {
        const child = definedType.type.accept(this);
        return `${TYPE_PREFIX}[${definedType.name}](${child})`;
    }
    visitError(error) {
        return `${ERROR_PREFIX}[${error.name}]`;
    }
    visitTypeArray(typeArray) {
        const child = typeArray.itemType.accept(this);
        return `array(${typeArray.size};${child})`;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        return `link(${typeDefinedLink.definedType})`;
    }
    visitTypeEnum(typeEnum) {
        const children = typeEnum.variants.map((variant) => variant.accept(this));
        return `enum[${typeEnum.name}](${children.join(',')})`;
    }
    visitTypeEnumEmptyVariant(typeEnumEmptyVariant) {
        return typeEnumEmptyVariant.name;
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        const child = typeEnumStructVariant.struct.accept(this);
        return `${typeEnumStructVariant.name}:${child}`;
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        const child = typeEnumTupleVariant.tuple.accept(this);
        return `${typeEnumTupleVariant.name}:${child}`;
    }
    visitTypeLeaf(typeLeaf) {
        return typeLeaf.type;
    }
    visitTypeLeafWrapper(typeLeafWrapper) {
        const child = typeLeafWrapper.leaf.accept(this);
        const { wrapper } = typeLeafWrapper;
        switch (wrapper.kind) {
            case 'DateTime':
                return `DateTime(${child})`;
            case 'Amount':
                return `Amount(${child},${wrapper.identifier},${wrapper.decimals})`;
            default:
                return child;
        }
    }
    visitTypeMap(typeMap) {
        const key = typeMap.keyType.accept(this);
        const value = typeMap.valueType.accept(this);
        return `map(${key},${value})`;
    }
    visitTypeOption(typeOption) {
        const child = typeOption.type.accept(this);
        return `option(${child})`;
    }
    visitTypeSet(typeSet) {
        const child = typeSet.type.accept(this);
        return `set(${child})`;
    }
    visitTypeStruct(typeStruct) {
        const children = typeStruct.fields.map((field) => field.accept(this));
        return `struct[${typeStruct.name}](${children.join(',')})`;
    }
    visitTypeStructField(typeStructField) {
        const child = typeStructField.type.accept(this);
        return `${typeStructField.name}:${child}`;
    }
    visitTypeTuple(typeTuple) {
        const children = typeTuple.itemTypes.map((itemType) => itemType.accept(this));
        return `tuple(${children.join(',')})`;
    }
    visitTypeVec(typeVec) {
        const child = typeVec.itemType.accept(this);
        return `vec(${child})`;
    }
}
exports.GetNodeInlineStringVisitor = GetNodeInlineStringVisitor;
//# sourceMappingURL=GetNodeInlineStringVisitor.js.map